#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016-2018, Blue Brain Project
#                          Raphael Dumusc <raphael.dumusc@epfl.ch>
#                          Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#                          Cyrille Favreau <cyrille.favreau@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.

"""Client that connects to a remote running Brayns instance which provides the supported API."""

import base64
import io

from PIL import Image
from .api_generator import build_api
from .rpcclient import RpcClient
from .utils import in_notebook, HTTP_METHOD_GET, HTTP_STATUS_OK, SCHEMA_ENDPOINT
from .version import MINIMAL_VERSION
from . import utils


def _obtain_registry(url):
    """Obtain the registry of exposed objects and RPCs from Brayns."""
    status = utils.http_request(HTTP_METHOD_GET, url, 'registry')
    if status.code != HTTP_STATUS_OK:
        raise Exception('Failed to obtain registry from Brayns')
    return status.contents


class Client(RpcClient):
    """Client that connects to a remote running Brayns instance which provides the supported API."""

    def __init__(self, url):
        """
        Create a new client instance by connecting to the given URL.

        :param str url: a string 'hostname:port' to connect to a running brayns instance
        """
        super(Client, self).__init__(url)
        self._check_version()
        self._build_api()

        if in_notebook():
            self._add_widgets()  # pragma: no cover

    def __str__(self):
        """Return a pretty-print on the currently connected Brayns instance."""
        # pylint: disable=E1101
        version = 'unknown'
        if self.version:
            version = '.'.join(str(x) for x in [self.version.major, self.version.minor,
                                                self.version.patch])
        return "Brayns version {0} running on {1}".format(version, self.url())

    # pylint: disable=W0613,W0622,E1101
    def image(self, size, format='jpg', animation_parameters=None, camera=None, quality=None,
              renderer=None, samples_per_pixel=None):
        """
        Request a snapshot from Brayns and return a PIL image.

        :param tuple size: (width,height) of the resulting image
        :param str format: image type as recognized by FreeImage
        :param object animation_parameters: animation params to use instead of current params
        :param object camera: camera to use instead of current camera
        :param int quality: compression quality between 1 (worst) and 100 (best)
        :param object renderer: renderer to use instead of current renderer
        :param int samples_per_pixel: samples per pixel to increase render quality
        :return: the PIL image of the current rendering, None on error obtaining the image
        :rtype: :py:class:`~PIL.Image.Image`
        """
        args = locals()
        del args['self']
        result = self.snapshot(response_timeout=None, **{k: v for k, v in args.items() if v})

        # error case: invalid request/parameters
        if 'code' in result:
            print(result['message'])
            return None

        # https://stackoverflow.com/a/9807138
        data = result['data']
        missing_padding = len(data) % 4
        if missing_padding != 0:
            data += b'=' * (4 - missing_padding)
        return Image.open(io.BytesIO(base64.b64decode(data)))

    def set_colormap(self, colormap='magma', colormap_size=256, intensity=1, opacity=1,
                     data_range=(0, 256)):
        """
        Set a colormap to Brayns.

        :param str colormap: color palette to use from matplotlib and seaborn
        :param int colormap_size: the number of colors to use to control precision
        :param int intensity: value to amplify the color values
        :param int opacity: opacity for colormap values
        :param typle data_range: data range on which values the colormap should be applied
        """
        import seaborn as sns
        palette = sns.color_palette(colormap, colormap_size)
        palette_size = len(palette)
        contributions = []
        diffuses = []
        # pylint: disable=E1101
        tf = self.transfer_function
        for i in range(0, palette_size):
            color = palette[i]
            diffuses.append([intensity * color[0], intensity * color[1], intensity * color[2],
                             opacity])
            contributions.append(0)
        tf.diffuse = diffuses
        tf.contribution = contributions
        tf.range = data_range
        tf.commit()

    def open_ui(self):
        """Open the Brayns UI in a new page of the default system browser."""
        import webbrowser
        url = 'https://bbp-brayns.epfl.ch?host=' + self.url()
        webbrowser.open(url)

    def _check_version(self):
        """Check if the Brayns' version is sufficient enough."""
        status = utils.http_request(HTTP_METHOD_GET, self.url(), 'version')
        if status.code != HTTP_STATUS_OK:
            raise Exception('Cannot obtain version from Brayns')

        version = '.'.join(str(x) for x in [status.contents['major'], status.contents['minor'],
                                            status.contents['patch']])

        import semver
        if semver.match(version, '<{0}'.format(MINIMAL_VERSION)):
            raise Exception('Brayns does not satisfy minimal required version; '
                            'needed {0}, got {1}'.format(MINIMAL_VERSION, version))

    def _build_api(self):
        """Fetch the registry and all schemas from the remote running Brayns to build the API."""
        registry = _obtain_registry(self.url())
        endpoints = {x.replace(SCHEMA_ENDPOINT, '') for x in registry}

        # batch request all schemas from all endpoints
        params = list()
        for endpoint in endpoints:
            params.append({'endpoint': endpoint})
        methods = ['schema']*len(params)
        schemas = self.batch_request(methods, params)

        schemas_dict = dict()
        for param, schema in zip(params, schemas):
            schemas_dict[param['endpoint']] = schema
        build_api(self, registry, schemas_dict)

    def _add_widgets(self):  # pragma: no cover
        """Add functions to the Brayns object to provide widgets for appropriate properties."""
        self._add_show_function()
        self._add_animation_slider()

    def _add_show_function(self):  # pragma: no cover
        """Add show() function for live streaming."""
        def function_builder():
            """Wrapper for returning the show() function."""
            def show():
                """Show the live rendering of Brayns."""
                self._setup_websocket()

                # pylint: disable=F0401,E1101
                from IPython.display import display
                import ipywidgets as widgets
                image = widgets.Image(format='jpg')
                image.value = base64.b64decode(self.image_jpeg()['data'])
                display(image)

                def image_update(data=None, close=False):
                    """Update callback on new image or when the websocket was closed."""
                    if close:
                        image.close()
                    elif data is not None:
                        image.value = data

                self._update_callback['image-jpeg'] = image_update

            return show

        setattr(self, 'show', function_builder())

    def _add_animation_slider(self):  # pragma: no cover
        """Add animation_slider() function for animation_parameters control."""
        def function_builder():
            """Wrapper for returning the animation_slider() function."""
            def animation_slider():
                """.Show slider to control animation"""
                self._setup_websocket()

                # pylint: disable=F0401,E1101
                import ipywidgets as widgets
                from IPython.display import display

                button = widgets.ToggleButton(value=self.animation_parameters.delta != 0,
                                              icon='play' if self.animation_parameters.delta == 0
                                              else 'pause')
                button.layout = widgets.Layout(width='40px')

                def on_button_change(change):
                    """Callback after play/pause button update to send delta for animation."""
                    button.icon = 'pause' if change['new'] else 'play'
                    self.set_animation_parameters(delta=1 if change['new'] else 0)

                button.observe(on_button_change, names='value')

                slider = widgets.IntSlider(min=self.animation_parameters.start,
                                           max=self.animation_parameters.end,
                                           value=self.animation_parameters.current)

                def on_value_change(change):
                    """Callback after slider update to send current for animation."""
                    self.set_animation_parameters(current=change['new'])

                slider.observe(on_value_change, names='value')

                w = widgets.HBox([button, slider])
                display(w)

                def slider_update(data=None, close=False):
                    """Update callback on update or when the websocket was closed."""
                    if close:
                        w.close()
                    elif data is not None:
                        slider.unobserve(on_value_change, names='value')
                        slider.min = data.start
                        slider.max = data.end
                        slider.value = data.current
                        slider.observe(on_value_change, names='value')

                self._update_callback['set-animation-parameters'] = slider_update

            return animation_slider

        setattr(self, 'animation_slider', function_builder())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016-2018, Blue Brain Project
#                          Raphael Dumusc <raphael.dumusc@epfl.ch>
#                          Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#                          Cyrille Favreau <cyrille.favreau@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.

"""The Brayns python client library provides access to remote control a running Brayns instance."""

from .client import Client
from .rpcclient import RpcClient
from .version import VERSION as __version__

__all__ = ['Client', 'RpcClient']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016-2018, Blue Brain Project
#                          Raphael Dumusc <raphael.dumusc@epfl.ch>
#                          Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#                          Cyrille Favreau <cyrille.favreau@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.

"""
Provides functionality to dynamically add properties, types and functions to a given object.

The information for the API generation is based on a registry and JSON-schema of exposed properties
and RPCs.
"""

import os
import python_jsonschema_objects as pjs
import inflection

from .utils import HTTP_METHOD_GET, HTTP_METHOD_PUT, HTTP_STATUS_OK, SCHEMA_ENDPOINT

from . import utils


def build_api(target_object, registry, schemas):
    """
    Add the API (types, properties, methods) to the given object from registry and schemas.

    :param object target_object: The target object where to add all generated properties, RPCs and
                                 types to
    :param dict registry: The registry of all endpoints which are properties and methods
    :param dict schemas: The schemas matching the endpoints from the registry
    """
    for registry_entry in registry.keys():
        if registry_entry in schemas:
            schema = schemas[registry_entry]
        elif registry_entry.endswith(SCHEMA_ENDPOINT):
            method_name = registry_entry[:-len(SCHEMA_ENDPOINT)]
            schema = schemas[method_name]

        if 'type' not in schema:
            continue

        if 'method' in schema['type']:
            _add_method(target_object, schema)
            continue
        if registry_entry.endswith(SCHEMA_ENDPOINT):
            continue

        writeable = HTTP_METHOD_PUT in registry[registry_entry]
        _try_add_property(target_object, registry_entry, schema, writeable)


def _try_add_property(target_object, registry_entry, schema, writeable):
    """
    Try to add registry_entry as a property.

    This will add a property with the name registry_entry which initializes itself with current
    value from Brayns on first access. Furthermore, it adds potential enum values as string
    constants and adds a commit() function if the property is writeable.

    :param object target_object: The target object where to add the property to
    :param str registry_entry: registry endpoint, e.g. foo[/schema]
    :param dict schema: The schema of the object behind the endpoint.
    :param bool writeable: if the property is writeable or not
    """
    classes = pjs.ObjectBuilder(schema).build_classes()
    class_names = dir(classes)

    # find the generated class name that matches our schema root title
    for c in class_names:
        if c.lower() == schema['title'].lower():
            class_name = c
            break

    class_type = getattr(classes, class_name)

    is_object = schema['type'] == 'object'
    if is_object:
        value = class_type()
        _add_enums(value, target_object)
        if writeable:
            _add_commit(target_object, class_type, registry_entry)
    else:  # array
        value = class_type(())

    # add member and property to target_object
    member = '_' + utils.underscorize(os.path.basename(registry_entry))
    setattr(target_object, member, value)
    _add_property(target_object, member, registry_entry, schema['type'])


def _create_method_with_object_parameter(param, method, description):
    """
    Create code for a method where each property of the param object is a key-value argument.

    :param dict param: the parameter object from the RPC
    :param str method: the name of RPC
    :param str description: the description of RPC
    :return: the code of the function
    :rtype: str
    """
    required = param['required'] if 'required' in param else list()
    optional = list()
    for s in param['properties'].keys():
        if s not in required:
            optional.append(s + '=None')
    arg_list = ', '.join(required)
    if optional and required:
        arg_list += ', '
    arg_list += ', '.join(filter(None, optional))
    return '''
        def function(self, {0}, response_timeout=5):
            """
            {1}
            """
            args = locals()
            del args['self']
            del args['response_timeout']
            return self.request("{2}", params={{k:v for k,v in args.items()
                                             if v is not None}},
                                             response_timeout=response_timeout)
        '''.format(arg_list, description, method)


def _create_method_with_array_parameter(name, method, description):
    """
    Create code for a method where the parameter is an array argument.

    :param str name: the name of the array argument
    :param str method: the name of RPC
    :param str description: the description of RPC
    :return: the code of the function
    :rtype: str
    """
    return '''
        def function(self, {1}, response_timeout=5):
            """
            {0}
            {2}
            """
            return self.request("{3}", params={1},
                                    response_timeout=response_timeout)
        '''.format(description, name, ":param {0}: {1}".format(name, description), method)


def _create_method_with_oneof_parameter(target_object, param, method, description):
    """
    Create code for a method where the parameter is from the oneOf array.

    Also create a type for each oneOf type and add it to target_object.

    :param object target_object: The target object where to add the oneOf types to
    :param list param: the oneOf array
    :param str method: name of RPC
    :param str description: description of RPC
    :return: the code of the function
    :rtype: str
    """
    param_types = list()
    for o in param:
        classes = pjs.ObjectBuilder(o).build_classes()
        class_names = dir(classes)

        # find the generated class name that matches our schema root title
        class_name = None
        for c in class_names:
            if c.lower() == inflection.camelize(o['title'].replace(' ', '_')).lower():
                class_name = c
                break

        # create class name <Type><Method w/o set->, e.g. Perspective+CameraParams
        pretty_class_name = utils.underscorize(method[4:])
        pretty_class_name = inflection.camelize(pretty_class_name)
        pretty_class_name = class_name + pretty_class_name

        # add type to target_object
        class_type = getattr(classes, class_name)
        setattr(target_object, pretty_class_name, class_type)
        param_types.append(pretty_class_name)

        # create and add potential enums to type
        _add_enums(class_type(), class_type)

    return '''
        def function(self, params, response_timeout=5):
            """
            {0}
            ":param: one of the params for the active type: {1}
            """
            return self.request("{2}", params.for_json(),
                                    response_timeout=response_timeout)
        '''.format(description, ', '.join(param_types), method)


def _add_enums(root_object, target_object):
    """
    Look for enums in the given object to create string constants <ENUM_CLASSNAME>_<ENUM_VALUE>.

    :param dict root_object: the object which may contain enums
    :param object target_object: The target object where to add the string constants to
    """
    for i in root_object.keys():
        enum = None
        propinfo = root_object.propinfo(i)
        if 'enum' in propinfo:
            enum = propinfo
        if propinfo['type'] == 'array':
            if 'enum' in propinfo['items']:
                enum = propinfo['items']
        if not enum:
            continue

        enum_class = str(i).upper()
        if 'title' in enum:
            enum_class = str(inflection.underscore(enum['title'])).upper()
        enum_class += "_"
        for val in enum['enum']:
            enum_value = enum_class + inflection.parameterize(val, '_').upper()
            setattr(target_object, enum_value, val)


def _add_method(target_object, schema):
    """
    Add a new method to target_object from the given schema that describes an RPC.

    :param object target_object: The target object where to add the method to
    :param dict schema: schema containing name, description, params of the RPC
    :raises Exception: if the param type of the RPC does not match oneOf, object or array
    """
    method = schema['title']
    func_name = str(utils.underscorize(os.path.basename(method)))

    if 'params' in schema and len(schema['params']) > 1:
        print("Multiple parameters for RPC '{0}' not supported".format(method))
        return

    description = schema['description']
    if 'params' in schema and len(schema['params']) == 1:
        params = schema['params'][0]
        if 'oneOf' in params:
            code = _create_method_with_oneof_parameter(target_object, params['oneOf'], method,
                                                       description)
        # in the absence of multiple parameters support, create a function with multiple
        # parameters from object properties
        elif params['type'] == 'object':
            code = _create_method_with_object_parameter(params, method, description)
        elif params['type'] == 'array':
            code = _create_method_with_array_parameter(params['name'], method, description)
        else:
            raise Exception('Invalid parameter type for method "{0}":'.format(method) +
                            ' must be "object", "array" or "oneOf"')
    else:
        code = '''
            def function(self, response_timeout=5):
                """
                {0}
                """
                return self.request("{1}", response_timeout=response_timeout)
            '''.format(description, method)

    d = {}
    exec(code.strip(), d)  # pylint: disable=W0122
    function = d['function']
    function.__name__ = func_name
    setattr(target_object.__class__, function.__name__, function)


def _add_commit(rpc_client, property_type, object_name):
    """Add commit() for given property type."""
    def commit_builder(url):
        """Wrapper for returning the property.commit() function."""
        def commit(prop):
            """Update the property in the remote side."""
            return rpc_client.request('set-' + os.path.basename(url), prop.as_dict())

        return commit

    setattr(property_type, 'commit', commit_builder(object_name))


def _add_property(target_object, member, property_name, property_type):
    """Add property to target_object which initializes itself on first access."""
    def getter_builder(member, property_name):
        """Wrapper for returning the property state."""
        def function(self):
            """Returns the current state for the property."""
            value = getattr(self, member)

            # Initialize on first access; updates are received via websocket
            if property_type == 'array':
                has_value = value.data
            else:
                has_value = value.as_dict()
            if not has_value or not target_object.connected():
                status = utils.http_request(HTTP_METHOD_GET, self.url(), property_name)
                if status.code == HTTP_STATUS_OK:
                    if property_type == 'array':
                        value.__init__(status.contents)
                    else:
                        value.__init__(**status.contents)

            if property_type == 'array':
                return value.data
            return value

        return function

    endpoint_name = os.path.basename(property_name)
    snake_case_name = utils.underscorize(endpoint_name)
    setattr(type(target_object), snake_case_name,
            property(fget=getter_builder(member, property_name),
                     doc='Access to the {0} property'.format(endpoint_name)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016-2018, Blue Brain Project
#                          Raphael Dumusc <raphael.dumusc@epfl.ch>
#                          Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#                          Cyrille Favreau <cyrille.favreau@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.

"""
The RpcClient manages a websocket connection to handle messaging with a remote Brayns instance.

It runs in a thread and provides methods to send notifications and requests in JSON-RPC format.
"""

import json
import threading
import time
import websocket

from .utils import set_http_protocol, set_ws_protocol, WS_PATH, underscorize

JSON_RPC_VERSION = '2.0'

try:
    import numpy as _np

    class Encoder(json.JSONEncoder):
        """Extends the default JSONEncoder to support numpy arrays"""

        def default(self, o):  # pylint: disable=E0202
            """Encoder for numpy arrays, for other types call super"""
            if isinstance(o, _np.ndarray):
                return o.tolist()
            return json.JSONEncoder.default(self, o)  # pragma: no cover
except ImportError:  # pragma: no cover
    Encoder = json.JSONEncoder


class RpcClient(object):
    """
    The RpcClient manages a websocket connection to handle messaging with a remote Brayns instance.

    It runs in a thread and provides methods to send notifications and requests in JSON-RPC format.
    """

    def __init__(self, url):
        """
        Initialize the RpcClient, but don't setup the websocket connection yet.

        Convert the URL to a proper format and initialize the state of the client. Does not
        establish the websocket connection yet. This will be postponed to either the first notify
        or request.

        :param str url: The address of the remote running Brayns instance.
        """
        self._url = set_http_protocol(url) + '/'

        self._ws = None
        self._ws_connected = False
        self._request_id = 0
        self._update_callback = {}
        self._ws_requests = {}

    def url(self):
        """
        Returns the address of the remote running Brayns instance.

        :return: The address of the remote running Brayns instance.
        :rtype: str
        """
        return self._url

    def connected(self):
        """Returns true if the websocket is connected to the remote Brayns instance."""
        return self._ws_connected

    def request(self, method, params=None, response_timeout=5):  # pragma: no cover
        """
        Invoke an RPC on the remote running Brayns instance and wait for its reponse.

        :param str method: name of the method to invoke
        :param dict params: params for the method
        :param int response_timeout: number of seconds to wait for the response
        :return: result or error of RPC
        :rtype: dict
        :raises Exception: if request was not answered within given response_timeout
        """
        request, result = self._make_request(method, params)

        self._setup_websocket()
        self._ws.send(json.dumps(request, cls=Encoder))

        if response_timeout:
            timeout = response_timeout * 10

            while not result['done'] and timeout:
                time.sleep(0.1)
                timeout -= 1

            if 'done' not in result:
                raise Exception('Request was not answered within {0} seconds'
                                .format(response_timeout))
        else:
            while not result['done']:
                time.sleep(0.0001)

        return result['result']

    def batch_request(self, methods, params, response_timeout=5):  # pragma: no cover
        """
        Invoke a batch RPC on the remote running Brayns instance and wait for its reponse.

        :param list methods: name of the methods to invoke
        :param list params: params for the methods
        :param int response_timeout: number of seconds to wait for the response
        :return: list of responses and/or errors of RPC
        :rtype: list
        :raises Exception: if methods and/or params are not a list
        :raises Exception: if request was not answered within given response_timeout
        """
        if not isinstance(methods, list) and not isinstance(params, list):
            raise Exception('Not a list of methods')

        requests = list()
        responses = list()
        for method, param in zip(methods, params):
            request, response = self._make_request(method, param)
            requests.append(request)
            responses.append(response)

        self._setup_websocket()
        self._ws.send(json.dumps(requests, cls=Encoder))

        result = responses[0]
        if response_timeout:
            timeout = response_timeout * 10

            while not result['done'] and timeout:
                time.sleep(0.1)
                timeout -= 1

            if 'done' not in result:
                raise Exception('Request was not answered within {0} seconds'
                                .format(response_timeout))
        else:
            while not result['done']:
                time.sleep(0.0001)

        results = list()
        for response in responses:
            results.append(response['result'])

        return results

    def notify(self, method, params=None):  # pragma: no cover
        """
        Invoke an RPC on the remote running Brayns instance without waiting for a response.

        :param str method: name of the method to invoke
        :param str params: params for the method
        """
        data = dict()
        data['jsonrpc'] = JSON_RPC_VERSION
        data['method'] = method
        if params:
            data['params'] = params

        self._setup_websocket()
        self._ws.send(json.dumps(data, cls=Encoder))

    def _make_request(self, method, params=None):  # pragma: no cover
        """
        Create a request object with given method and params and setup the response callback.

        :param str method: name of the method to invoke
        :param str params: params for the method
        :return: request and response dict
        :rtype: dict
        """
        request = dict()
        request['jsonrpc'] = JSON_RPC_VERSION
        request['id'] = self._request_id
        request['method'] = method
        if params:
            request['params'] = params

        response = {'done': False, 'result': None}

        def callback(payload):
            """
            The callback for the response.

            :param dict payload: the actual response data
            """
            response['result'] = payload
            response['done'] = True

        self._ws_requests[self._request_id] = callback
        self._request_id += 1

        return request, response

    def _setup_websocket(self):  # pragma: no cover
        """
        Setups websocket to handle binary (image) and text (all properties) messages.

        The websocket app runs in a separate thread to unblock all notebook cells.
        """
        if self._ws_connected:
            return

        def on_open(ws):
            # pylint: disable=unused-argument
            """Websocket is open, remember this state."""
            self._ws_connected = True

        def on_data(ws, data, data_type, cont):
            # pylint: disable=unused-argument
            """Websocket received data, handle it."""
            if data_type == websocket.ABNF.OPCODE_TEXT:
                data = json.loads(data)

                # check if the received data is a response from a previous request
                if self._handle_response(data):
                    return

                # if the received data is a change notification from any known object
                if not data['method'].startswith('set-'):
                    return

                # remove the 'set-' part of the data method to find the property to update its
                # content
                prop_name = underscorize(data['method'][4:])
                prop = getattr(self, '_' + prop_name, None)
                if prop is None:
                    return

                prop.__init__(**data['params'])

                # call an optionally registered callback on the newly updated property
                if data['method'] in self._update_callback:
                    self._update_callback[data['method']](data=prop)
            elif data_type == websocket.ABNF.OPCODE_BINARY:
                if 'image-jpeg' in self._update_callback:
                    self._update_callback['image-jpeg'](data=data)

        def on_close(ws):
            # pylint: disable=unused-argument
            """Websocket is closing, notify all registered callbacks to e.g. close widgets."""
            self._ws_connected = False
            for f in self._update_callback.values():
                f(close=True)
            self._update_callback.clear()

        websocket.enableTrace(False)

        self._ws = websocket.WebSocketApp(set_ws_protocol(self._url) + WS_PATH,
                                          subprotocols=['rockets'], on_open=on_open,
                                          on_data=on_data, on_close=on_close)
        ws_thread = threading.Thread(target=self._ws.run_forever)
        ws_thread.daemon = True
        ws_thread.start()

        conn_timeout = 5
        while not self._ws_connected and conn_timeout:
            time.sleep(0.2)
            conn_timeout -= 1

    def _handle_response(self, data):  # pragma: no cover
        """
        Handle a potential JSON-RPC response message.

        :param dict data: data of the reply, either a dict or a list (batch request)
        :return: True if a request was handled, False otherwise
        :rtype: bool
        """
        if 'id' not in data and not isinstance(data, list):
            return False

        if isinstance(data, list):
            for response in data:
                self._finish_request(response)
        else:
            self._finish_request(data)

        return True

    def _finish_request(self, data):  # pragma: no cover
        """
        Extract payload from data which can be result or error and invoke the response callback.

        :param dict data: data of the reply
        """
        payload = None
        if 'result' in data:
            payload = None if data['result'] == '' or data['result'] == 'OK' else data['result']
        elif 'error' in data:
            payload = data['error']

        if data['id'] in self._ws_requests:
            self._ws_requests[data['id']](payload)
            self._ws_requests.pop(data['id'])
        else:
            if 'id' in data and data['id']:
                print('Got error response for request ' + str(data['id']) + ': ' + str(payload))
            else:
                print('Got error response: ' + str(payload))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016-2018, Blue Brain Project
#                          Raphael Dumusc <raphael.dumusc@epfl.ch>
#                          Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#                          Cyrille Favreau <cyrille.favreau@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.

"""Utils module for http request and notebook stuff."""

import sys
from collections import OrderedDict
import requests


HTTP_METHOD_PUT = 'PUT'
HTTP_METHOD_GET = 'GET'
HTTP_METHOD_DELETE = 'DELETE'
HTTP_METHOD_POST = 'POST'
HTTP_STATUS_OK = 200

HTTP_PREFIX = 'http://'
HTTPS_PREFIX = 'https://'
WS_PREFIX = 'ws://'
WSS_PREFIX = 'wss://'

WS_PATH = '/ws'

SCHEMA_ENDPOINT = '/schema'


class Status(object):
    """Holds the execution status of an HTTP request."""

    def __init__(self, code, contents):
        """Initialize status object with given code and contents."""
        self.code = code
        self.contents = contents


# pylint: disable=R0912
def http_request(method, url, command, body=None, query_params=None):  # pragma: no cover
    """
    Perform http requests to the given URL and return the applications' response.

    :param str method: the type of HTTP request, PUT or GET are supported
    :param str url: the URL of the applications' http server
    :param str body: optional body for PUT requests
    :param str command: the type of HTTP command to be executed on the target app
    :param str query_params: the query params to append to the request url
    :return: JSON-encoded response of the request
    :rtype: Status
    :raises Exception: on connection error
    """
    full_url = url
    request = None
    full_url += command
    try:
        if method == HTTP_METHOD_POST:
            if body == '':
                request = requests.post(full_url, params=query_params)
            else:
                request = requests.post(full_url, data=body, params=query_params)
        elif method == HTTP_METHOD_PUT:
            if body == '':
                request = requests.put(full_url, params=query_params)
            else:
                request = requests.put(full_url, data=body, params=query_params)
        elif method == HTTP_METHOD_GET:
            request = requests.get(full_url, params=query_params)
            if request.status_code == 502:
                raise requests.exceptions.ConnectionError('Bad Gateway 502')
        elif method == HTTP_METHOD_DELETE:
            if body == '':
                request = requests.delete(full_url, params=query_params)
            else:
                request = requests.delete(full_url, data=body, params=query_params)
        js = ''
        if request.content:
            if request.status_code == 200:
                js = request.json(object_pairs_hook=OrderedDict)
            else:
                js = request.text
        response = Status(request.status_code, js)
        request.close()
    except requests.exceptions.ConnectionError:
        raise Exception('ERROR: Failed to connect to Brayns, did you start it with the '
                        '--http-server command line option?')
    return response


def in_notebook():
    """
    Check we are running in a notebook or not.

    :return: ``True`` if the module is running in IPython kernel, ``False`` if in IPython shell or
    other Python shell.
    :rtype: bool
    """
    return 'ipykernel' in sys.modules


def set_http_protocol(url):
    """
    Set the http protocol to the url if it is not present.

    :param str url: Url to be checked
    :return: Url with protocol
    :rtype: str
    """
    if url.find(HTTP_PREFIX) == -1 and url.find(HTTPS_PREFIX) == -1:
        return HTTP_PREFIX + url
    return url


def set_ws_protocol(url):
    """
    Set the WebSocket protocol according to the resource url.

    :param str url: Url to be checked
    :return: Url preprend with ws for http, wss for https
    :rtype: str
    """
    if url.find(HTTPS_PREFIX) != -1:
        return WSS_PREFIX + url[len(HTTPS_PREFIX):]
    if url.find(HTTP_PREFIX) != -1:
        return WS_PREFIX + url[len(HTTP_PREFIX):]
    return WS_PREFIX + url


def underscorize(word):
    """
    Opposite of inflection.dasherize; replace dashes with underscore in the word.

    :param str word: the word to underscorize
    :return: the underscorized word
    :rtype: str
    """
    return word.replace('-', '_')
